#!/usr/bin/env python

def setUp():
    print "setUp"
    pass

def tearDown():
    print "tearDown"
    pass

def test_003():
    print "003"
    pass

def test_002():
    print "002"
    pass

def test_001():
    print "001"
    pass

print "hello"
